@@ -1,5 +1,5 @@
 /*-
- * Copyright (c) 2009 Michihiro NAKAJIMA
+ * Copyright (c) 2009-2011 Michihiro NAKAJIMA
  * Copyright (c) 2003-2007 Kees Zeelenberg
  * All rights reserved.
  *
@@ -45,20 +45,24 @@
  */
 
 #if defined(_WIN32) && !defined(__CYGWIN__)
-#define _WIN32_WINNT 0x0500
-#define WINVER       0x0500
 
 #include "archive_platform.h"
+#include "archive_private.h"
+#include <ctype.h>
 #include <errno.h>
 #include <stddef.h>
+#ifdef HAVE_SYS_UTIME_H
 #include <sys/utime.h>
+#endif
 #include <sys/stat.h>
+#include <locale.h>
 #include <process.h>
 #include <stdlib.h>
 #include <wchar.h>
 #include <windows.h>
+#include <share.h>
 
-#define EPOC_TIME	(116444736000000000ULL)
+#define EPOC_TIME ARCHIVE_LITERAL_ULL(116444736000000000)
 
 struct ustat {
 	int64_t		st_atime;
@@ -78,9 +82,6 @@ struct ustat {
 	dev_t		st_rdev;
 };
 
-/* Local replacement for undocumented Windows CRT function. */
-static void la_dosmaperr(unsigned long e);
-
 /* Transform 64-bits ino into 32-bits by hashing.
  * You do not forget that really unique number size is 64-bits.
  */
@@ -89,7 +90,6 @@ static __inline ino_t
 getino(struct ustat *ub)
 {
 	ULARGE_INTEGER ino64;
-
 	ino64.QuadPart = ub->st_ino;
 	/* I don't know this hashing is correct way */
 	return (ino64.LowPart ^ (ino64.LowPart >> INOSIZE));
@@ -101,52 +101,58 @@ getino(struct ustat *ub)
  * characters.
  * see also http://msdn.microsoft.com/en-us/library/aa365247.aspx
  */
-static wchar_t *
-permissive_name(const char *name)
+wchar_t *
+__la_win_permissive_name(const char *name)
+{
+	wchar_t *wn;
+	wchar_t *ws;
+	size_t ll;
+
+	ll = strlen(name);
+	wn = malloc((ll + 1) * sizeof(wchar_t));
+	if (wn == NULL)
+		return (NULL);
+	ll = mbstowcs(wn, name, ll);
+	if (ll == (size_t)-1) {
+		free(wn);
+		return (NULL);
+	}
+	wn[ll] = L'\0';
+	ws = __la_win_permissive_name_w(wn);
+	free(wn);
+	return (ws);
+}
+
+wchar_t *
+__la_win_permissive_name_w(const wchar_t *wname)
 {
 	wchar_t *wn, *wnp;
 	wchar_t *ws, *wsp;
-	size_t l, len, slen;
+	DWORD l, len, slen;
 	int unc;
 
-	len = strlen(name);
-	wn = malloc((len + 1) * sizeof(wchar_t));
-	if (wn == NULL)
+	/* Get a full-pathname. */
+	l = GetFullPathNameW(wname, 0, NULL, NULL);
+	if (l == 0)
 		return (NULL);
-	l = MultiByteToWideChar(CP_ACP, 0, name, len, wn, len);
-	if (l == 0) {
-		free(wn);
-		return (NULL);
-	}
-	wn[l] = L'\0';
-
-	/* Get a full path names */
-	l = GetFullPathNameW(wn, 0, NULL, NULL);
-	if (l == 0) {
-		free(wn);
-		return (NULL);
-	}
 	wnp = malloc(l * sizeof(wchar_t));
-	if (wnp == NULL) {
-		free(wn);
+	if (wnp == NULL)
 		return (NULL);
-	}
-	len = GetFullPathNameW(wn, l, wnp, NULL);
-	free(wn);
+	len = GetFullPathNameW(wname, l, wnp, NULL);
 	wn = wnp;
 
 	if (wnp[0] == L'\\' && wnp[1] == L'\\' &&
 	    wnp[2] == L'?' && wnp[3] == L'\\')
-		/* We have already permissive names. */
+		/* We have already a permissive name. */
 		return (wn);
 
 	if (wnp[0] == L'\\' && wnp[1] == L'\\' &&
 		wnp[2] == L'.' && wnp[3] == L'\\') {
-		/* Device names */
+		/* This is a device name */
 		if (((wnp[4] >= L'a' && wnp[4] <= L'z') ||
 		     (wnp[4] >= L'A' && wnp[4] <= L'Z')) &&
 		    wnp[5] == L':' && wnp[6] == L'\\')
-			wnp[2] = L'?';/* Not device names. */
+			wnp[2] = L'?';/* Not device name. */
 		return (wn);
 	}
 
@@ -188,11 +194,16 @@ permissive_name(const char *name)
 		wsp += 4;
 		slen -= 4;
 	}
-	wcsncpy_s(wsp, slen, wnp, _TRUNCATE);
+	wcsncpy(wsp, wnp, slen);
+	wsp[slen - 1] = L'\0'; /* Ensure null termination. */
 	free(wn);
 	return (ws);
 }
 
+/*
+ * Create a file handle.
+ * This can exceed MAX_PATH limitation.
+ */
 static HANDLE
 la_CreateFile(const char *path, DWORD dwDesiredAccess, DWORD dwShareMode,
     LPSECURITY_ATTRIBUTES lpSecurityAttributes, DWORD dwCreationDisposition,
@@ -208,7 +219,7 @@ la_CreateFile(const char *path, DWORD dwDesiredAccess, DWORD dwShareMode,
 		return (handle);
 	if (GetLastError() != ERROR_PATH_NOT_FOUND)
 		return (handle);
-	wpath = permissive_name(path);
+	wpath = __la_win_permissive_name(path);
 	if (wpath == NULL)
 		return (handle);
 	handle = CreateFileW(wpath, dwDesiredAccess, dwShareMode,
@@ -218,331 +229,11 @@ la_CreateFile(const char *path, DWORD dwDesiredAccess, DWORD dwShareMode,
 	return (handle);
 }
 
-static size_t
-wequallen(const wchar_t *s1, const wchar_t *s2)
-{
-	size_t i = 0;
-
-	while (*s1 != L'\0' && *s2 != L'\0' && *s1 == *s2) {
-		++s1; ++s2; ++i;
-	}
-	return (i);
-}
-
-/* Check that path1 and path2 can be hard-linked by each other.
- * Both arguments must be made by permissive_name function. 
- */
-static int
-canHardLinkW(const wchar_t *path1, const wchar_t *path2)
-{
-	wchar_t root[MAX_PATH];
-	wchar_t fs[32];
-	const wchar_t *s;
-	int r;
-
-	r = wequallen(path1, path2);
-	/* Is volume-name the same? */
-	if (r < 7)
-		return (0);
-	if (wcsncmp(path1, L"\\\\?\\UNC\\", 8) == 0) {
-		int len;
-
-		s = path1 + 8;
-		if (*s == L'\\')
-			return (0);
-		/*         012345678
-		 * Name : "\\?\UNC\Server\Share\"
-		 *                       ^ search
-		 */
-		s = wcschr(++s, L'\\');
-		if (s == NULL)
-			return (0);
-		if (*++s == L'\\')
-			return (0);
-		/*         012345678
-		 * Name : "\\?\UNC\Server\Share\"
-		 *                             ^ search
-		 */
-		s = wcschr(++s, L'\\');
-		if (s == NULL)
-			return (0);
-		s++;
-		/*         012345678
-		 * Name : "\\?\UNC\Server\Share\xxxx"
-		 *                 ^--- len ----^
-		 */
-		len = (int)(s - path1 - 8);
-		/* Is volume-name the same? */
-		if (r < len + 8)
-			return (0);
-		/* Is volume-name too long? */
-		if (sizeof(root) -3 < len)
-			return (0);
-		root[0] = root[1] = L'\\';
-		wcsncpy(root + 2, path1 + 8 , len);
-		/* root : "\\Server\Share\" */
-		root[2 + len] = L'\0';
-	} else if (wcsncmp(path1, L"\\\\?\\", 4) == 0) {
-		s = path1 + 4;
-		if ((!iswalpha(*s)) || s[1] != L':' || s[2] != L'\\')
-			return (0);
-		wcsncpy(root, path1 + 4, 3);
-		root[3] = L'\0';
-	} else
-		return (0);
-	if (!GetVolumeInformationW(root, NULL, 0, NULL, NULL, NULL, fs, sizeof(fs)))
-		return (0);
-	if (wcscmp(fs, L"NTFS") == 0)
-		return (1);
-	else
-		return (0);
-}
-
-/* Make a link to src called dst.  */
-static int
-__link(const char *src, const char *dst, int sym)
-{
-	wchar_t *wsrc, *wdst;
-	int res, retval;
-	DWORD attr;
-
-	if (src == NULL || dst == NULL) {
-		set_errno (EINVAL);
-		return -1;
-	}
-
-	wsrc = permissive_name(src);
-	wdst = permissive_name(dst);
-	if (wsrc == NULL || wdst == NULL) {
-		if (wsrc != NULL)
-			free(wsrc);
-		if (wdst != NULL)
-			free(wdst);
-		set_errno (EINVAL);
-		return -1;
-	}
-
-	if ((attr = GetFileAttributesW(wsrc)) != -1) {
-		if ((attr & FILE_ATTRIBUTE_DIRECTORY) != 0) {
-			errno = EPERM;
-			retval = -1;
-			goto exit;
-		}
-		if (!sym && canHardLinkW(wsrc, wdst))
-			res = CreateHardLinkW(wdst, wsrc, NULL);
-		else
-			res = CopyFileW(wsrc, wdst, FALSE);
-	} else {
-		/* wsrc does not exist; try src prepend it with the dirname of wdst */
-		wchar_t *wnewsrc, *slash;
-		int i, n, slen, wlen;
-
-		if (strlen(src) >= 3 && isalpha((unsigned char)src[0]) &&
-		    src[1] == ':' && src[2] == '\\') {
-			/* Original src name is already full-path */
-			retval = -1;
-			goto exit;
-		}
-		if (src[0] == '\\') {
-			/* Original src name is almost full-path
-			 * (maybe src name is without drive) */
-			retval = -1;
-			goto exit;
-		}
-
-		wnewsrc = malloc ((wcslen(wsrc) + wcslen(wdst) + 1) * sizeof(wchar_t));
-		if (wnewsrc == NULL) {
-			errno = ENOMEM;
-			retval = -1;
-			goto exit;
-		}
-		/* Copying a dirname of wdst */
-		wcscpy(wnewsrc, wdst);
-		slash = wcsrchr(wnewsrc, L'\\');
-		if (slash != NULL)
-			*++slash = L'\0';
-		else
-			wcscat(wnewsrc, L"\\");
-		/* Converting multi-byte src to wide-char src */
-		wlen = wcslen(wsrc);
-		slen = strlen(src);
-		n = MultiByteToWideChar(CP_ACP, 0, src, slen, wsrc, slen);
-		if (n == 0) {
-			free (wnewsrc);
-			retval = -1;
-			goto exit;
-		}
-		for (i = 0; i < n; i++)
-			if (wsrc[i] == L'/')
-				wsrc[i] = L'\\';
-		wcsncat(wnewsrc, wsrc, n);
-		/* Check again */
-		attr = GetFileAttributesW(wnewsrc);
-		if (attr == -1 || (attr & FILE_ATTRIBUTE_DIRECTORY) != 0) {
-			if (attr == -1)
-				la_dosmaperr(GetLastError());
-			else
-				errno = EPERM;
-			free (wnewsrc);
-			retval = -1;
-			goto exit;
-		}
-		if (!sym && canHardLinkW(wnewsrc, wdst))
-			res = CreateHardLinkW(wdst, wnewsrc, NULL);
-		else
-			res = CopyFileW(wnewsrc, wdst, FALSE);
-		free (wnewsrc);
-	}
-	if (res == 0) {
-		la_dosmaperr(GetLastError());
-		retval = -1;
-	} else
-		retval = 0;
-exit:
-	free(wsrc);
-	free(wdst);
-	return (retval);
-}
-
-/* Make a hard link to src called dst.  */
-int
-link(const char *src, const char *dst)
-{
-	return __link (src, dst, 0);
-}
-
-/* Make a symbolic link to FROM called TO.  */
-int symlink (from, to)
-     const char *from;
-     const char *to;
-{
-	return __link (from, to, 1);
-}
-
-int
-ftruncate(int fd, off_t length)
-{
-	LARGE_INTEGER distance;
-	HANDLE handle;
-
-	if (fd < 0) {
-		errno = EBADF;
-		return (-1);
-	}
-	handle = (HANDLE)_get_osfhandle(fd);
-	if (GetFileType(handle) != FILE_TYPE_DISK) {
-		errno = EBADF;
-		return (-1);
-	}
-	distance.QuadPart = length;
-	if (!SetFilePointerEx(handle, distance, NULL, FILE_BEGIN)) {
-		la_dosmaperr(GetLastError());
-		return (-1);
-	}
-	if (!SetEndOfFile(handle)) {
-		la_dosmaperr(GetLastError());
-		return (-1);
-	}
-	return (0);
-}
-
-#define WINTIME(sec, usec)	((Int32x32To64(sec, 10000000) + EPOC_TIME) + (usec * 10))
-static int
-__hutimes(HANDLE handle, const struct __timeval *times)
-{
-	ULARGE_INTEGER wintm;
-	FILETIME fatime, fmtime;
-
-	wintm.QuadPart = WINTIME(times[0].tv_sec, times[0].tv_usec);
-	fatime.dwLowDateTime = wintm.LowPart;
-	fatime.dwHighDateTime = wintm.HighPart;
-	wintm.QuadPart = WINTIME(times[1].tv_sec, times[1].tv_usec);
-	fmtime.dwLowDateTime = wintm.LowPart;
-	fmtime.dwHighDateTime = wintm.HighPart;
-	if (SetFileTime(handle, NULL, &fatime, &fmtime) == 0) {
-		errno = EINVAL;
-		return (-1);
-	}
-	return (0);
-}
-
-int
-futimes(int fd, const struct __timeval *times)
-{
-
-	return (__hutimes((HANDLE)_get_osfhandle(fd), times));
-}
-
-int
-utimes(const char *name, const struct __timeval *times)
-{
-	int ret;
-	HANDLE handle;
-
-	handle = la_CreateFile(name, GENERIC_READ | GENERIC_WRITE,
-	    FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING,
-	    FILE_FLAG_BACKUP_SEMANTICS, NULL);
-	if (handle == INVALID_HANDLE_VALUE) {
-		la_dosmaperr(GetLastError());
-		return (-1);
-	}
-	ret = __hutimes(handle, times);
-	CloseHandle(handle);
-	return (ret);
-}
-
-int
-la_chdir(const char *path)
-{
-	wchar_t *ws;
-	int r;
-
-	r = SetCurrentDirectoryA(path);
-	if (r == 0) {
-		if (GetLastError() != ERROR_FILE_NOT_FOUND) {
-			la_dosmaperr(GetLastError());
-			return (-1);
-		}
-	} else
-		return (0);
-	ws = permissive_name(path);
-	if (ws == NULL) {
-		errno = EINVAL;
-		return (-1);
-	}
-	r = SetCurrentDirectoryW(ws);
-	free(ws);
-	if (r == 0) {
-		la_dosmaperr(GetLastError());
-		return (-1);
-	}
-	return (0);
-}
-
-int
-la_chmod(const char *path, mode_t mode)
-{
-	wchar_t *ws;
-	int r;
-
-	r = _chmod(path, mode);
-	if (r >= 0 || errno != ENOENT)
-		return (r);
-	ws = permissive_name(path);
-	if (ws == NULL) {
-		errno = EINVAL;
-		return (-1);
-	}
-	r = _wchmod(ws, mode);
-	free(ws);
-	return (r);
-}
-
 /*
- * This fcntl is limited implemention.
+ * This fcntl is limited implementation.
  */
 int
-la_fcntl(int fd, int cmd, int val)
+__la_fcntl(int fd, int cmd, int val)
 {
 	HANDLE handle;
 
@@ -559,85 +250,9 @@ la_fcntl(int fd, int cmd, int val)
 	return (-1);
 }
 
-__int64
-la_lseek(int fd, __int64 offset, int whence)
-{
-	LARGE_INTEGER distance;
-	LARGE_INTEGER newpointer;
-	HANDLE handle;
-
-	if (fd < 0) {
-		errno = EBADF;
-		return (-1);
-	}
-	handle = (HANDLE)_get_osfhandle(fd);
-	if (GetFileType(handle) != FILE_TYPE_DISK) {
-		errno = EBADF;
-		return (-1);
-	}
-	distance.QuadPart = offset;
-	if (!SetFilePointerEx(handle, distance, &newpointer, whence)) {
-		DWORD lasterr;
-
-		lasterr = GetLastError();
-		if (lasterr == ERROR_BROKEN_PIPE)
-			return (0);
-		if (lasterr == ERROR_ACCESS_DENIED)
-			errno = EBADF;
-		else
-			la_dosmaperr(lasterr);
-		return (-1);
-	}
-	return (newpointer.QuadPart);
-}
-
+/* This can exceed MAX_PATH limitation. */
 int
-la_mkdir(const char *path, mode_t mode)
-{
-	wchar_t *ws;
-	int r;
-
-	(void)mode;/* UNUSED */
-	r = CreateDirectoryA(path, NULL);
-	if (r == 0) {
-		DWORD lasterr = GetLastError();
-		if (lasterr != ERROR_FILENAME_EXCED_RANGE &&
-			lasterr != ERROR_PATH_NOT_FOUND) {
-			la_dosmaperr(GetLastError());
-			return (-1);
-		}
-	} else
-		return (0);
-	ws = permissive_name(path);
-	if (ws == NULL) {
-		errno = EINVAL;
-		return (-1);
-	}
-	r = CreateDirectoryW(ws, NULL);
-	free(ws);
-	if (r == 0) {
-		la_dosmaperr(GetLastError());
-		return (-1);
-	}
-	return (0);
-}
-
-/* Windows' mbstowcs is differrent error handling from other unix mbstowcs.
- * That one is using MultiByteToWideChar function with MB_PRECOMPOSED and
- * MB_ERR_INVALID_CHARS flags.
- * This implements for only to pass libarchive_test.
- */
-size_t
-la_mbstowcs(wchar_t *wcstr, const char *mbstr, size_t nwchars)
-{
-
-	return (MultiByteToWideChar(CP_ACP, MB_ERR_INVALID_CHARS,
-	    mbstr, (int)strlen(mbstr), wcstr,
-	    (int)nwchars));
-}
-
-int
-la_open(const char *path, int flags, ...)
+__la_open(const char *path, int flags, ...)
 {
 	va_list ap;
 	wchar_t *ws;
@@ -654,15 +269,15 @@ la_open(const char *path, int flags, ...)
 		 * "Permission denied" error.
 		 */
 		attr = GetFileAttributesA(path);
-		if (attr == -1 && GetLastError() == ERROR_PATH_NOT_FOUND) {
-			ws = permissive_name(path);
+		if (attr == (DWORD)-1 && GetLastError() == ERROR_PATH_NOT_FOUND) {
+			ws = __la_win_permissive_name(path);
 			if (ws == NULL) {
 				errno = EINVAL;
 				return (-1);
 			}
 			attr = GetFileAttributesW(ws);
 		}
-		if (attr == -1) {
+		if (attr == (DWORD)-1) {
 			la_dosmaperr(GetLastError());
 			free(ws);
 			return (-1);
@@ -692,11 +307,17 @@ la_open(const char *path, int flags, ...)
 		}
 	}
 	if (ws == NULL) {
+#if defined(__BORLANDC__)
+		/* Borland has no mode argument.
+		   TODO: Fix mode of new file.  */
+		r = _open(path, flags);
+#else
 		r = _open(path, flags, pmode);
+#endif
 		if (r < 0 && errno == EACCES && (flags & O_CREAT) != 0) {
-			/* simular other POSIX system action to pass a test */
+			/* Simulate other POSIX system action to pass our test suite. */
 			attr = GetFileAttributesA(path);
-			if (attr == -1)
+			if (attr == (DWORD)-1)
 				la_dosmaperr(GetLastError());
 			else if (attr & FILE_ATTRIBUTE_DIRECTORY)
 				errno = EISDIR;
@@ -706,7 +327,7 @@ la_open(const char *path, int flags, ...)
 		}
 		if (r >= 0 || errno != ENOENT)
 			return (r);
-		ws = permissive_name(path);
+		ws = __la_win_permissive_name(path);
 		if (ws == NULL) {
 			errno = EINVAL;
 			return (-1);
@@ -714,9 +335,9 @@ la_open(const char *path, int flags, ...)
 	}
 	r = _wopen(ws, flags, pmode);
 	if (r < 0 && errno == EACCES && (flags & O_CREAT) != 0) {
-		/* simular other POSIX system action to pass a test */
+		/* Simulate other POSIX system action to pass our test suite. */
 		attr = GetFileAttributesW(ws);
-		if (attr == -1)
+		if (attr == (DWORD)-1)
 			la_dosmaperr(GetLastError());
 		else if (attr & FILE_ATTRIBUTE_DIRECTORY)
 			errno = EISDIR;
@@ -728,7 +349,7 @@ la_open(const char *path, int flags, ...)
 }
 
 ssize_t
-la_read(int fd, void *buf, size_t nbytes)
+__la_read(int fd, void *buf, size_t nbytes)
 {
 	HANDLE handle;
 	DWORD bytes_read, lasterr;
@@ -742,6 +363,10 @@ la_read(int fd, void *buf, size_t nbytes)
 		errno = EBADF;
 		return (-1);
 	}
+	/* Do not pass 0 to third parameter of ReadFile(), read bytes.
+	 * This will not return to application side. */
+	if (nbytes == 0)
+		return (0);
 	handle = (HANDLE)_get_osfhandle(fd);
 	if (GetFileType(handle) == FILE_TYPE_PIPE) {
 		DWORD sta;
@@ -778,26 +403,6 @@ la_read(int fd, void *buf, size_t nbytes)
 	return ((ssize_t)bytes_read);
 }
 
-/* Remove directory */
-int
-la_rmdir(const char *path)
-{
-	wchar_t *ws;
-	int r;
-
-	r = _rmdir(path);
-	if (r >= 0 || errno != ENOENT)
-		return (r);
-	ws = permissive_name(path);
-	if (ws == NULL) {
-		errno = EINVAL;
-		return (-1);
-	}
-	r = _wrmdir(ws);
-	free(ws);
-	return (r);
-}
-
 /* Convert Windows FILETIME to UTC */
 __inline static void
 fileTimeToUTC(const FILETIME *filetime, time_t *time, long *ns)
@@ -817,9 +422,9 @@ fileTimeToUTC(const FILETIME *filetime, time_t *time, long *ns)
 }
 
 /* Stat by handle
- * Windows' stat() does not accept path which is added "\\?\" especially "?"
+ * Windows' stat() does not accept the path added "\\?\" especially "?"
  * character.
- * It means we cannot access a long name path(which is longer than MAX_PATH).
+ * It means we cannot access the long name path longer than MAX_PATH.
  * So I've implemented simular Windows' stat() to access the long name path.
  * And I've added some feature.
  * 1. set st_ino by nFileIndexHigh and nFileIndexLow of
@@ -908,7 +513,7 @@ __hstat(HANDLE handle, struct ustat *st)
 	st->st_nlink = 1;
 	st->st_dev = 0;
 #else
-	/* Getting FileIndex as i-node. We have to remove a sequence which
+	/* Getting FileIndex as i-node. We should remove a sequence which
 	 * is high-16-bits of nFileIndexHigh. */
 	ino64.HighPart = info.nFileIndexHigh & 0x0000FFFFUL;
 	ino64.LowPart  = info.nFileIndexLow;
@@ -940,8 +545,13 @@ copy_stat(struct stat *st, struct ustat *us)
 	st->st_rdev = us->st_rdev;
 }
 
+/*
+ * TODO: Remove a use of __la_fstat and __la_stat.
+ * We should use GetFileInformationByHandle in place
+ * where We still use the *stat functions.
+ */
 int
-la_fstat(int fd, struct stat *st)
+__la_fstat(int fd, struct stat *st)
 {
 	struct ustat u;
 	int ret;
@@ -961,15 +571,16 @@ la_fstat(int fd, struct stat *st)
 	return (ret);
 }
 
+/* This can exceed MAX_PATH limitation. */
 int
-la_stat(const char *path, struct stat *st)
+__la_stat(const char *path, struct stat *st)
 {
 	HANDLE handle;
 	struct ustat u;
 	int ret;
 
 	handle = la_CreateFile(path, 0, 0, NULL, OPEN_EXISTING,
-		FILE_FLAG_BACKUP_SEMANTICS | FILE_ATTRIBUTE_READONLY,
+		FILE_FLAG_BACKUP_SEMANTICS,
 		NULL);
 	if (handle == INVALID_HANDLE_VALUE) {
 		la_dosmaperr(GetLastError());
@@ -998,36 +609,18 @@ la_stat(const char *path, struct stat *st)
 	return (ret);
 }
 
-int
-la_unlink(const char *path)
-{
-	wchar_t *ws;
-	int r;
-
-	r = _unlink(path);
-	if (r >= 0 || errno != ENOENT)
-		return (r);
-	ws = permissive_name(path);
-	if (ws == NULL) {
-		errno = EINVAL;
-		return (-1);
-	}
-	r = _wunlink(ws);
-	free(ws);
-	return (r);
-}
-
 /*
- * This waitpid is limited implemention.
+ * This waitpid is limited implementation.
  */
 pid_t
-la_waitpid(pid_t wpid, int *status, int option)
+__la_waitpid(pid_t wpid, int *status, int option)
 {
 	HANDLE child;
 	DWORD cs, ret;
 
 	(void)option;/* UNUSED */
-	child = OpenProcess(PROCESS_ALL_ACCESS, FALSE, wpid);
+	*status = 0;
+	child = OpenProcess(PROCESS_QUERY_INFORMATION | SYNCHRONIZE, FALSE, wpid);
 	if (child == NULL) {
 		la_dosmaperr(GetLastError());
 		return (-1);
@@ -1052,9 +645,9 @@ la_waitpid(pid_t wpid, int *status, int option)
 }
 
 ssize_t
-la_write(int fd, const void *buf, size_t nbytes)
+__la_write(int fd, const void *buf, size_t nbytes)
 {
-	uint32_t bytes_written;
+	DWORD bytes_written;
 
 #ifdef _WIN64
 	if (nbytes > UINT32_MAX)
@@ -1173,8 +766,8 @@ static const struct {
 	{	ERROR_NOT_ENOUGH_QUOTA, ENOMEM	}
 };
 
-static void
-la_dosmaperr(unsigned long e)
+void
+__la_dosmaperr(unsigned long e)
 {
 	int			i;
 
@@ -1198,103 +791,4 @@ la_dosmaperr(unsigned long e)
 	return;
 }
 
-#if !defined(HAVE_OPENSSL_MD5_H) && !defined(HAVE_OPENSSL_SHA_H)
-/*
- * Message digest functions.
- */
-static void
-Digest_Init(Digest_CTX *ctx, ALG_ID algId)
-{
-
-	ctx->valid = 0;
-	if (!CryptAcquireContext(&ctx->cryptProv, NULL, NULL,
-	    PROV_RSA_FULL, 0)) {
-		if (GetLastError() != NTE_BAD_KEYSET)
-			return;
-		if (!CryptAcquireContext(&ctx->cryptProv, NULL, NULL,
-		    PROV_RSA_FULL, CRYPT_NEWKEYSET))
-			return;
-	}
-
-	if (!CryptCreateHash(ctx->cryptProv, algId, 0, 0, &ctx->hash)) {
-		CryptReleaseContext(ctx->cryptProv, 0);
-		return;
-	}
-
-	ctx->valid = 1;
-}
-
-static void
-Digest_Update(Digest_CTX *ctx, const unsigned char *buf, size_t len)
-{
-
-	if (!ctx->valid)
-	return;
-
-	CryptHashData(ctx->hash, buf, (DWORD)len, 0);
-}
-
-static void
-Digest_Final(unsigned char *buf, int bufsize, Digest_CTX *ctx)
-{
-	DWORD siglen = bufsize;
-
-	if (!ctx->valid)
-		return;
-
-	CryptGetHashParam(ctx->hash, HP_HASHVAL, buf, &siglen, 0);
-	CryptDestroyHash(ctx->hash);
-	CryptReleaseContext(ctx->cryptProv, 0);
-	ctx->valid = 0;
-}
-
-#define DIGEST_INIT(name, algid) \
-void name ## _Init(Digest_CTX *ctx)\
-{\
-	Digest_Init(ctx, algid);\
-}
-
-#define DIGEST_UPDATE(name) \
-void name ## _Update(Digest_CTX *ctx, const unsigned char *buf, size_t len)\
-{\
-	Digest_Update(ctx, buf, len);\
-}
-
-#define DIGEST_FINAL(name, size) \
-void name ## _Final(unsigned char buf[size], Digest_CTX *ctx)\
-{\
-	Digest_Final(buf, size, ctx);\
-}
-
-DIGEST_INIT(MD5, CALG_MD5)
-DIGEST_UPDATE(MD5)
-DIGEST_FINAL(MD5, MD5_DIGEST_LENGTH)
-
-DIGEST_INIT(SHA1, CALG_SHA1)
-DIGEST_UPDATE(SHA1)
-DIGEST_FINAL(SHA1, SHA1_DIGEST_LENGTH)
-
-/*
- * SHA256 nor SHA384 nor SHA512 are not supported on Windows XP and Windows 2000.
- */
-#ifdef CALG_SHA256
-DIGEST_INIT(SHA256, CALG_SHA256)
-DIGEST_UPDATE(SHA256)
-DIGEST_FINAL(SHA256, SHA256_DIGEST_LENGTH)
-#endif
-
-#ifdef CALG_SHA384
-DIGEST_INIT(SHA384, CALG_SHA384)
-DIGEST_UPDATE(SHA384)
-DIGEST_FINAL(SHA384, SHA384_DIGEST_LENGTH)
-#endif
-
-#ifdef CALG_SHA512
-DIGEST_INIT(SHA512, CALG_SHA512)
-DIGEST_UPDATE(SHA512)
-DIGEST_FINAL(SHA512, SHA384_DIGEST_LENGTH)
-#endif
-
-#endif /* !HAVE_OPENSSL_MD5_H && !HAVE_OPENSSL_SHA_H */
-
 #endif /* _WIN32 && !__CYGWIN__ */
