@@ -40,12 +40,20 @@ DEFINE_TEST(test_0)
 	struct stat st;
 
 	failure("File %s does not exist?!", testprog);
-	if (!assertEqualInt(0, stat(testprog, &st)))
+	if (!assertEqualInt(0, stat(testprogfile, &st))) {
+		fprintf(stderr,
+		    "\nFile %s does not exist; aborting test.\n\n",
+		    testprog);
 		exit(1);
+	}
 
 	failure("%s is not executable?!", testprog);
-	if (!assert((st.st_mode & 0111) != 0))
+	if (!assert((st.st_mode & 0111) != 0)) {
+		fprintf(stderr,
+		    "\nFile %s not executable; aborting test.\n\n",
+		    testprog);
 		exit(1);
+	}
 
 	/*
 	 * Try to succesfully run the program; this requires that
