@@ -18,6 +18,7 @@
  * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
  * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+
  * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
  * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
@@ -48,6 +49,9 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/tree.c,v 1.9 2008/11/27 05:49:52 kientzle Ex
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
+#ifdef HAVE_DIRECT_H
+#include <direct.h>
+#endif
 #ifdef HAVE_DIRENT_H
 #include <dirent.h>
 #endif
@@ -66,6 +70,9 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/tree.c,v 1.9 2008/11/27 05:49:52 kientzle Ex
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#if defined(HAVE_WINDOWS_H) && !defined(__CYGWIN__)
+#include <windows.h>
+#endif
 
 #include "tree.h"
 
@@ -76,27 +83,38 @@ __FBSDID("$FreeBSD: src/usr.bin/tar/tree.c,v 1.9 2008/11/27 05:49:52 kientzle Ex
  */
 
 struct tree_entry {
+	int depth;
 	struct tree_entry *next;
 	struct tree_entry *parent;
 	char *name;
 	size_t dirname_length;
 	dev_t dev;
 	ino_t ino;
+	int flags;
+	/* How to return back to the parent of a symlink. */
 #ifdef HAVE_FCHDIR
-	int fd;
+	int symlink_parent_fd;
 #elif defined(_WIN32) && !defined(__CYGWIN__)
-	char *fullpath;
+	char *symlink_parent_path;
 #else
 #error fchdir function required.
 #endif
-	int flags;
 };
 
 /* Definitions for tree_entry.flags bitmap. */
 #define	isDir 1 /* This entry is a regular directory. */
 #define	isDirLink 2 /* This entry is a symbolic link to a directory. */
-#define	needsPreVisit 4 /* This entry needs to be previsited. */
-#define	needsPostVisit 8 /* This entry needs to be postvisited. */
+#define	needsFirstVisit 4 /* This is an initial entry. */
+#define	needsDescent 8 /* This entry needs to be previsited. */
+#define	needsOpen 16 /* This is a directory that needs to be opened. */
+#define	needsAscent 32 /* This entry needs to be postvisited. */
+
+/*
+ * On Windows, "first visit" is handled as a pattern to be handed to
+ * _findfirst().  This is consistent with Windows conventions that
+ * file patterns are handled within the application.  On Posix,
+ * "first visit" is just returned to the client.
+ */
 
 /*
  * Local data for this package.
@@ -104,21 +122,28 @@ struct tree_entry {
 struct tree {
 	struct tree_entry	*stack;
 	struct tree_entry	*current;
+#if defined(_WIN32) && !defined(__CYGWIN__)
+	HANDLE d;
+	BY_HANDLE_FILE_INFORMATION fileInfo;
+#define	INVALID_DIR_HANDLE INVALID_HANDLE_VALUE
+	WIN32_FIND_DATA _findData;
+	WIN32_FIND_DATA *findData;
+#else
 	DIR	*d;
-#ifdef HAVE_FCHDIR
-	int	 initialDirFd;
-#elif defined(_WIN32) && !defined(__CYGWIN__)
-	char	*initialDir;
+#define	INVALID_DIR_HANDLE NULL
+	struct dirent *de;
 #endif
 	int	 flags;
 	int	 visit_type;
 	int	 tree_errno; /* Error code from last failed operation. */
 
+	/* Dynamically-sized buffer for holding path */
 	char	*buff;
-	const char	*basename;
 	size_t	 buff_length;
-	size_t	 path_length;
-	size_t	 dirname_length;
+
+	const char *basename; /* Last path element */
+	size_t	 dirname_length; /* Leading dir length */
+	size_t	 path_length; /* Total path length */
 
 	int	 depth;
 	int	 openCount;
@@ -129,37 +154,51 @@ struct tree {
 };
 
 /* Definitions for tree.flags bitmap. */
-#define needsReturn 8  /* Marks first entry as not having been returned yet. */
-#define hasStat 16  /* The st entry is set. */
-#define hasLstat 32 /* The lst entry is set. */
+#define	hasStat 16  /* The st entry is valid. */
+#define	hasLstat 32 /* The lst entry is valid. */
+#define	hasFileInfo 64 /* The Windows fileInfo entry is valid. */
 
+#if defined(_WIN32) && !defined(__CYGWIN__)
+static int
+tree_dir_next_windows(struct tree *t, const char *pattern);
+#else
+static int
+tree_dir_next_posix(struct tree *t);
+#endif
 
 #ifdef HAVE_DIRENT_D_NAMLEN
 /* BSD extension; avoids need for a strlen() call. */
-#define D_NAMELEN(dp)	(dp)->d_namlen
+#define	D_NAMELEN(dp)	(dp)->d_namlen
 #else
-#define D_NAMELEN(dp)	(strlen((dp)->d_name))
+#define	D_NAMELEN(dp)	(strlen((dp)->d_name))
 #endif
 
-#if 0
 #include <stdio.h>
 void
 tree_dump(struct tree *t, FILE *out)
 {
+	char buff[300];
 	struct tree_entry *te;
 
 	fprintf(out, "\tdepth: %d\n", t->depth);
 	fprintf(out, "\tbuff: %s\n", t->buff);
-	fprintf(out, "\tpwd: "); fflush(stdout); system("pwd");
-	fprintf(out, "\taccess: %s\n", t->basename);
+	fprintf(out, "\tpwd: %s\n", getcwd(buff, sizeof(buff)));
+	fprintf(out, "\tbasename: %s\n", t->basename);
 	fprintf(out, "\tstack:\n");
 	for (te = t->stack; te != NULL; te = te->next) {
-		fprintf(out, "\t\tte->name: %s%s%s\n", te->name,
-		    te->flags & needsPreVisit ? "" : " *",
-		    t->current == te ? " (current)" : "");
+		fprintf(out, "\t\t%s%d:\"%s\" %s%s%s%s%s%s\n",
+		    t->current == te ? "*" : " ",
+		    te->depth,
+		    te->name,
+		    te->flags & needsFirstVisit ? "V" : "",
+		    te->flags & needsDescent ? "D" : "",
+		    te->flags & needsOpen ? "O" : "",
+		    te->flags & needsAscent ? "A" : "",
+		    te->flags & isDirLink ? "L" : "",
+		    (t->current == te && t->d) ? "+" : ""
+		);
 	}
 }
-#endif
 
 /*
  * Add a directory path to the current stack.
@@ -172,24 +211,29 @@ tree_push(struct tree *t, const char *path)
 	te = malloc(sizeof(*te));
 	memset(te, 0, sizeof(*te));
 	te->next = t->stack;
+	te->parent = t->current;
+	if (te->parent)
+		te->depth = te->parent->depth + 1;
 	t->stack = te;
 #ifdef HAVE_FCHDIR
-	te->fd = -1;
-#elif defined(_WIN32) && !defined(__CYGWIN__)
-	te->fullpath = NULL;
-#endif
+	te->symlink_parent_fd = -1;
 	te->name = strdup(path);
-	te->flags = needsPreVisit | needsPostVisit;
+#elif defined(_WIN32) && !defined(__CYGWIN__)
+	te->symlink_parent_path = NULL;
+	te->name = strdup(path);
+#endif
+	te->flags = needsDescent | needsOpen | needsAscent;
 	te->dirname_length = t->dirname_length;
 }
 
 /*
- * Append a name to the current path.
+ * Append a name to the current dir path.
  */
 static void
 tree_append(struct tree *t, const char *name, size_t name_length)
 {
 	char *p;
+	size_t size_needed;
 
 	if (t->buff != NULL)
 		t->buff[t->dirname_length] = '\0';
@@ -198,12 +242,16 @@ tree_append(struct tree *t, const char *name, size_t name_length)
 		name_length--;
 
 	/* Resize pathname buffer as needed. */
-	while (name_length + 1 + t->dirname_length >= t->buff_length) {
-		t->buff_length *= 2;
+	size_needed = name_length + 1 + t->dirname_length;
+	if (t->buff_length < size_needed) {
 		if (t->buff_length < 1024)
 			t->buff_length = 1024;
+		while (t->buff_length < size_needed)
+			t->buff_length *= 2;
 		t->buff = realloc(t->buff, t->buff_length);
 	}
+	if (t->buff == NULL)
+		abort();
 	p = t->buff + t->dirname_length;
 	t->path_length = t->dirname_length + name_length;
 	/* Add a separating '/' if it's needed. */
@@ -211,7 +259,11 @@ tree_append(struct tree *t, const char *name, size_t name_length)
 		*p++ = '/';
 		t->path_length ++;
 	}
+#if HAVE_STRNCPY_S
+	strncpy_s(p, t->buff_length - (p - t->buff), name, name_length);
+#else
 	strncpy(p, name, name_length);
+#endif
 	p[name_length] = '\0';
 	t->basename = p;
 }
@@ -222,24 +274,76 @@ tree_append(struct tree *t, const char *name, size_t name_length)
 struct tree *
 tree_open(const char *path)
 {
+#ifdef HAVE_FCHDIR
 	struct tree *t;
 
 	t = malloc(sizeof(*t));
 	memset(t, 0, sizeof(*t));
-	tree_append(t, path, strlen(path));
-#ifdef HAVE_FCHDIR
-	t->initialDirFd = open(".", O_RDONLY);
-#elif defined(_WIN32) && !defined(__CYGWIN__)
-	t->initialDir = getcwd(NULL, 0);
-#endif
-	/*
-	 * During most of the traversal, items are set up and then
-	 * returned immediately from tree_next().  That doesn't work
-	 * for the very first entry, so we set a flag for this special
-	 * case.
-	 */
-	t->flags = needsReturn;
+	/* First item is set up a lot like a symlink traversal. */
+	tree_push(t, path);
+	t->stack->flags = needsFirstVisit | isDirLink | needsAscent;
+	t->stack->symlink_parent_fd = open(".", O_RDONLY);
+	t->openCount++;
+	t->d = INVALID_DIR_HANDLE;
 	return (t);
+#elif defined(_WIN32) && !defined(__CYGWIN__)
+	struct tree *t;
+	char *cwd = _getcwd(NULL, 0);
+	char *pathname, *p, *base;
+	wchar_t *wcs, *wp;
+	size_t l, wlen;
+
+	/* Take care of '\' character in multi-byte character-set.
+	 * Some multi-byte character-set have been using '\' character
+	 * for a part of its character code. */
+	l = MultiByteToWideChar(CP_OEMCP, 0, path, strlen(path), NULL, 0);
+	if (l == 0)
+		abort();
+	wcs = malloc(sizeof(*wcs) * (l+1));
+	if (wcs == NULL)
+		abort();
+	l = MultiByteToWideChar(CP_OEMCP, 0, path, strlen(path), wcs, l);
+	wcs[l] = L'\0';
+	wlen = l;
+	for (wp = wcs; *wp != L'\0'; ++wp) {
+		if (*wp == L'\\')
+			*wp = L'/';
+	}
+	l = WideCharToMultiByte(CP_OEMCP, 0, wcs, wlen, NULL, 0, NULL, NULL);
+	if (l == 0)
+		abort();
+	pathname = malloc(l+1);
+	if (pathname == NULL)
+		abort();
+	l = WideCharToMultiByte(CP_OEMCP, 0, wcs, wlen, pathname, l, NULL, NULL);
+	pathname[l] = '\0';
+	free(wcs);
+	base = pathname;
+
+	t = malloc(sizeof(*t));
+	memset(t, 0, sizeof(*t));
+	/* First item is set up a lot like a symlink traversal. */
+	/* printf("Looking for wildcard in %s\n", path); */
+	/* TODO: wildcard detection here screws up on \\?\c:\ UNC names */
+	if (strchr(base, '*') || strchr(base, '?')) {
+		/* It has a wildcard in it... */
+		/* Separate the last element. */
+		p = strrchr(base, '/');
+		if (p != NULL) {
+			*p = '\0';
+			chdir(base);
+			tree_append(t, base, p - base);
+			t->dirname_length = t->path_length;
+			base = p + 1;
+		}
+	}
+	tree_push(t, base);
+	free(pathname);
+	t->stack->flags = needsFirstVisit | isDirLink | needsAscent;
+	t->stack->symlink_parent_path = cwd;
+	t->d = INVALID_DIR_HANDLE;
+	return (t);
+#endif
 }
 
 /*
@@ -255,22 +359,26 @@ tree_ascend(struct tree *t)
 	t->depth--;
 	if (te->flags & isDirLink) {
 #ifdef HAVE_FCHDIR
-		if (fchdir(te->fd) != 0) {
+		if (fchdir(te->symlink_parent_fd) != 0) {
 			t->tree_errno = errno;
 			r = TREE_ERROR_FATAL;
 		}
-		close(te->fd);
+		close(te->symlink_parent_fd);
 #elif defined(_WIN32) && !defined(__CYGWIN__)
-		if (chdir(te->fullpath) != 0) {
+		if (SetCurrentDirectory(te->symlink_parent_path) == 0) {
 			t->tree_errno = errno;
 			r = TREE_ERROR_FATAL;
 		}
-		free(te->fullpath);
-		te->fullpath = NULL;
+		free(te->symlink_parent_path);
+		te->symlink_parent_path = NULL;
 #endif
 		t->openCount--;
 	} else {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+		if (SetCurrentDirectory("..") == 0) {
+#else
 		if (chdir("..") != 0) {
+#endif
 			t->tree_errno = errno;
 			r = TREE_ERROR_FATAL;
 		}
@@ -286,16 +394,18 @@ tree_pop(struct tree *t)
 {
 	struct tree_entry *te;
 
-	t->buff[t->dirname_length] = '\0';
+	if (t->buff)
+		t->buff[t->dirname_length] = '\0';
 	if (t->stack == t->current && t->current != NULL)
 		t->current = t->current->parent;
 	te = t->stack;
 	t->stack = te->next;
 	t->dirname_length = te->dirname_length;
-	t->basename = t->buff + t->dirname_length;
-	/* Special case: starting dir doesn't skip leading '/'. */
-	if (t->dirname_length > 0)
-		t->basename++;
+	if (t->buff) {
+		t->basename = t->buff + t->dirname_length;
+		while (t->basename[0] == '/')
+			t->basename++;
+	}
 	free(te->name);
 	free(te);
 }
@@ -306,101 +416,179 @@ tree_pop(struct tree *t)
 int
 tree_next(struct tree *t)
 {
-	struct dirent *de = NULL;
 	int r;
 
 	/* If we're called again after a fatal error, that's an API
 	 * violation.  Just crash now. */
 	if (t->visit_type == TREE_ERROR_FATAL) {
-		const char *msg = "Unable to continue traversing"
-		    " directory heirarchy after a fatal error.";
-		write(2, msg, strlen(msg));
-		*(int *)0 = 1; /* Deliberate SEGV; NULL pointer dereference. */
-		exit(1); /* In case the SEGV didn't work. */
-	}
-
-	/* Handle the startup case by returning the initial entry. */
-	if (t->flags & needsReturn) {
-		t->flags &= ~needsReturn;
-		return (t->visit_type = TREE_REGULAR);
+		fprintf(stderr, "Unable to continue traversing"
+		    " directory hierarchy after a fatal error.");
+		abort();
 	}
 
 	while (t->stack != NULL) {
 		/* If there's an open dir, get the next entry from there. */
-		while (t->d != NULL) {
-			de = readdir(t->d);
-			if (de == NULL) {
-				closedir(t->d);
-				t->d = NULL;
-			} else if (de->d_name[0] == '.'
-			    && de->d_name[1] == '\0') {
-				/* Skip '.' */
-			} else if (de->d_name[0] == '.'
-			    && de->d_name[1] == '.'
-			    && de->d_name[2] == '\0') {
-				/* Skip '..' */
-			} else {
-				/*
-				 * Append the path to the current path
-				 * and return it.
-				 */
-				tree_append(t, de->d_name, D_NAMELEN(de));
-				t->flags &= ~hasLstat;
-				t->flags &= ~hasStat;
-				return (t->visit_type = TREE_REGULAR);
-			}
+		if (t->d != INVALID_DIR_HANDLE) {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+			r = tree_dir_next_windows(t, NULL);
+#else
+			r = tree_dir_next_posix(t);
+#endif
+			if (r == 0)
+				continue;
+			return (r);
 		}
 
-		/* If the current dir needs to be visited, set it up. */
-		if (t->stack->flags & needsPreVisit) {
+		if (t->stack->flags & needsFirstVisit) {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+			char *d = t->stack->name;
+			t->stack->flags &= ~needsFirstVisit;
+			if (strchr(d, '*') || strchr(d, '?')) {
+				r = tree_dir_next_windows(t, d);
+				if (r == 0)
+					continue;
+				return (r);
+			}
+			/* Not a pattern, handle it as-is... */
+#endif
+			/* Top stack item needs a regular visit. */
 			t->current = t->stack;
 			tree_append(t, t->stack->name, strlen(t->stack->name));
-			t->stack->flags &= ~needsPreVisit;
+			/* t->dirname_length = t->path_length; */
+			/* tree_pop(t); */
+			t->stack->flags &= ~needsFirstVisit;
+			return (t->visit_type = TREE_REGULAR);
+		} else if (t->stack->flags & needsDescent) {
+			/* Top stack item is dir to descend into. */
+			t->current = t->stack;
+			tree_append(t, t->stack->name, strlen(t->stack->name));
+			t->stack->flags &= ~needsDescent;
 			/* If it is a link, set up fd for the ascent. */
 			if (t->stack->flags & isDirLink) {
 #ifdef HAVE_FCHDIR
-				t->stack->fd = open(".", O_RDONLY);
-#elif defined(_WIN32) && !defined(__CYGWIN__)
-				t->stack->fullpath = getcwd(NULL, 0);
-#endif
+				t->stack->symlink_parent_fd = open(".", O_RDONLY);
 				t->openCount++;
 				if (t->openCount > t->maxOpenCount)
 					t->maxOpenCount = t->openCount;
+#elif defined(_WIN32) && !defined(__CYGWIN__)
+				t->stack->symlink_parent_path = _getcwd(NULL, 0);
+#endif
 			}
 			t->dirname_length = t->path_length;
-			if (chdir(t->stack->name) != 0) {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+			if (t->path_length == 259 || !SetCurrentDirectory(t->stack->name) != 0)
+#else
+			if (chdir(t->stack->name) != 0)
+#endif
+			{
 				/* chdir() failed; return error */
 				tree_pop(t);
 				t->tree_errno = errno;
 				return (t->visit_type = TREE_ERROR_DIR);
 			}
 			t->depth++;
-			t->d = opendir(".");
-			if (t->d == NULL) {
+			return (t->visit_type = TREE_POSTDESCENT);
+		} else if (t->stack->flags & needsOpen) {
+			t->stack->flags &= ~needsOpen;
+#if defined(_WIN32) && !defined(__CYGWIN__)
+			r = tree_dir_next_windows(t, "*");
+#else
+			r = tree_dir_next_posix(t);
+#endif
+			if (r == 0)
+				continue;
+			return (r);
+		} else if (t->stack->flags & needsAscent) {
+		        /* Top stack item is dir and we're done with it. */
+			r = tree_ascend(t);
+			tree_pop(t);
+			t->visit_type = r != 0 ? r : TREE_POSTASCENT;
+			return (t->visit_type);
+		} else {
+			/* Top item on stack is dead. */
+			tree_pop(t);
+			t->flags &= ~hasLstat;
+			t->flags &= ~hasStat;
+		}
+	}
+	return (t->visit_type = 0);
+}
+
+#if defined(_WIN32) && !defined(__CYGWIN__)
+static int
+tree_dir_next_windows(struct tree *t, const char *pattern)
+{
+	const char *name;
+	size_t namelen;
+	int r;
+
+	for (;;) {
+		if (pattern != NULL) {
+			t->d = FindFirstFile(pattern, &t->_findData);
+			if (t->d == INVALID_DIR_HANDLE) {
 				r = tree_ascend(t); /* Undo "chdir" */
 				tree_pop(t);
 				t->tree_errno = errno;
 				t->visit_type = r != 0 ? r : TREE_ERROR_DIR;
 				return (t->visit_type);
 			}
-			t->flags &= ~hasLstat;
-			t->flags &= ~hasStat;
-			t->basename = ".";
-			return (t->visit_type = TREE_POSTDESCENT);
+			t->findData = &t->_findData;
+			pattern = NULL;
+		} else if (!FindNextFile(t->d, &t->_findData)) {
+			FindClose(t->d);
+			t->d = INVALID_DIR_HANDLE;
+			t->findData = NULL;
+			return (0);
 		}
+		name = t->findData->cFileName;
+		namelen = strlen(name);
+		t->flags &= ~hasLstat;
+		t->flags &= ~hasStat;
+		if (name[0] == '.' && name[1] == '\0')
+			continue;
+		if (name[0] == '.' && name[1] == '.' && name[2] == '\0')
+			continue;
+		tree_append(t, name, namelen);
+		return (t->visit_type = TREE_REGULAR);
+	}
+}
+#else
+static int
+tree_dir_next_posix(struct tree *t)
+{
+	int r;
+	const char *name;
+	size_t namelen;
 
-		/* We've done everything necessary for the top stack entry. */
-		if (t->stack->flags & needsPostVisit) {
-			r = tree_ascend(t);
+	if (t->d == NULL) {
+		if ((t->d = opendir(".")) == NULL) {
+			r = tree_ascend(t); /* Undo "chdir" */
 			tree_pop(t);
-			t->flags &= ~hasLstat;
-			t->flags &= ~hasStat;
-			t->visit_type = r != 0 ? r : TREE_POSTASCENT;
+			t->tree_errno = errno;
+			t->visit_type = r != 0 ? r : TREE_ERROR_DIR;
 			return (t->visit_type);
 		}
 	}
-	return (t->visit_type = 0);
+	for (;;) {
+		t->de = readdir(t->d);
+		if (t->de == NULL) {
+			closedir(t->d);
+			t->d = INVALID_DIR_HANDLE;
+			return (0);
+		}
+		name = t->de->d_name;
+		namelen = D_NAMELEN(t->de);
+		t->flags &= ~hasLstat;
+		t->flags &= ~hasStat;
+		if (name[0] == '.' && name[1] == '\0')
+			continue;
+		if (name[0] == '.' && name[1] == '.' && name[2] == '\0')
+			continue;
+		tree_append(t, name, namelen);
+		return (t->visit_type = TREE_REGULAR);
+	}
 }
+#endif
 
 /*
  * Return error code.
@@ -437,25 +625,51 @@ const struct stat *
 tree_current_stat(struct tree *t)
 {
 	if (!(t->flags & hasStat)) {
-		if (stat(t->basename, &t->st) != 0)
+		if (stat(tree_current_access_path(t), &t->st) != 0)
 			return NULL;
 		t->flags |= hasStat;
 	}
 	return (&t->st);
 }
 
+#if defined(_WIN32) && !defined(__CYGWIN__)
+const BY_HANDLE_FILE_INFORMATION *
+tree_current_file_information(struct tree *t)
+{
+	if (!(t->flags & hasFileInfo)) {
+		HANDLE h = CreateFile(tree_current_access_path(t),
+			0, 0, NULL,
+			OPEN_EXISTING,
+			FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT,
+			NULL);
+		if (h == INVALID_HANDLE_VALUE)
+			return NULL;
+		if (!GetFileInformationByHandle(h, &t->fileInfo)) {
+			CloseHandle(h);
+			return NULL;
+		}
+		CloseHandle(h);
+		t->flags |= hasFileInfo;
+	}
+	return (&t->fileInfo);
+}
+#endif
 /*
  * Get the lstat() data for the entry just returned from tree_next().
  */
 const struct stat *
 tree_current_lstat(struct tree *t)
 {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+	return (tree_current_stat(t));
+#else
 	if (!(t->flags & hasLstat)) {
-		if (lstat(t->basename, &t->lst) != 0)
+		if (lstat(tree_current_access_path(t), &t->lst) != 0)
 			return NULL;
 		t->flags |= hasLstat;
 	}
 	return (&t->lst);
+#endif
 }
 
 /*
@@ -464,8 +678,14 @@ tree_current_lstat(struct tree *t)
 int
 tree_current_is_dir(struct tree *t)
 {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+	if (t->findData)
+		return (t->findData->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
+	if (tree_current_file_information(t))
+		return (t->fileInfo.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY);
+	return (0);
+#else
 	const struct stat *st;
-
 	/*
 	 * If we already have lstat() info, then try some
 	 * cheap tests to determine if this is a dir.
@@ -490,6 +710,7 @@ tree_current_is_dir(struct tree *t)
 		return 0;
 	/* Use the definitive test.  Hopefully this is cached. */
 	return (S_ISDIR(st->st_mode));
+#endif
 }
 
 /*
@@ -500,6 +721,11 @@ tree_current_is_dir(struct tree *t)
 int
 tree_current_is_physical_dir(struct tree *t)
 {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+	if (tree_current_is_physical_link(t))
+		return (0);
+	return (tree_current_is_dir(t));
+#else
 	const struct stat *st;
 
 	/*
@@ -523,6 +749,7 @@ tree_current_is_physical_dir(struct tree *t)
 		return 0;
 	/* Use the definitive test.  Hopefully this is cached. */
 	return (S_ISDIR(st->st_mode));
+#endif
 }
 
 /*
@@ -531,10 +758,21 @@ tree_current_is_physical_dir(struct tree *t)
 int
 tree_current_is_physical_link(struct tree *t)
 {
+#if defined(_WIN32) && !defined(__CYGWIN__)
+#ifndef IO_REPARSE_TAG_SYMLINK
+/* Old SDKs do not provide IO_REPARSE_TAG_SYMLINK */
+#define	IO_REPARSE_TAG_SYMLINK 0xA000000CL
+#endif
+	if (t->findData)
+		return ((t->findData->dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT)
+				&& (t->findData->dwReserved0 == IO_REPARSE_TAG_SYMLINK));
+	return (0);
+#else
 	const struct stat *st = tree_current_lstat(t);
 	if (st == NULL)
 		return 0;
 	return (S_ISLNK(st->st_mode));
+#endif
 }
 
 /*
@@ -582,21 +820,23 @@ tree_close(struct tree *t)
 	/* Release anything remaining in the stack. */
 	while (t->stack != NULL)
 		tree_pop(t);
-	if (t->buff)
-		free(t->buff);
-	/* chdir() back to where we started. */
+	free(t->buff);
+	/* TODO: Ensure that premature close() resets cwd */
+#if 0
 #ifdef HAVE_FCHDIR
 	if (t->initialDirFd >= 0) {
-		fchdir(t->initialDirFd);
+		int s = fchdir(t->initialDirFd);
+		(void)s; /* UNUSED */
 		close(t->initialDirFd);
 		t->initialDirFd = -1;
 	}
 #elif defined(_WIN32) && !defined(__CYGWIN__)
 	if (t->initialDir != NULL) {
-		chdir(t->initialDir);
+		SetCurrentDir(t->initialDir);
 		free(t->initialDir);
 		t->initialDir = NULL;
 	}
+#endif
 #endif
 	free(t);
 }
